from datetime import date, datetime
from typing import Optional, Union

from rcheck.checks.shared import (
    InvalidRuntype,
    _assert_simple_type_fail_message,
    _isinstance,
    type_name,
)

#
# General condition
#


def _assert_true_fail_message(name: str, message: Optional[str] = None):
    return InvalidRuntype(f"Conditions not met for {name}." + (f"\n\n{message}" if message is not None else ""))


def assert_cond(condition: bool, name: str, message: Optional[str] = None):
    if not condition:
        raise _assert_true_fail_message(name, message)


#
# bool
#


def is_bool(val: object):
    return _isinstance(val, bool)


def assert_bool(val: object, name: str, message: Optional[str] = None):
    if not is_bool(val):
        raise _assert_simple_type_fail_message("bool", val, name, message)


def is_opt_bool(val: object):
    return val is None or is_bool(val)


def assert_opt_bool(val: object, name: str, message: Optional[str] = None):
    if not is_opt_bool(val):
        raise _assert_simple_type_fail_message("Optional[bool]", val, name, message)


#
# str
#


def is_str(val: object):
    return _isinstance(val, str)


def assert_str(val: object, name: str, message: Optional[str] = None):
    if not is_str(val):
        raise _assert_simple_type_fail_message("str", val, name, message)


def is_opt_str(val: object):
    return val is None or is_str(val)


def assert_opt_str(val: object, name: str, message: Optional[str] = None):
    if not is_opt_str(val):
        raise _assert_simple_type_fail_message("Optional[str]", val, name, message)


#
# int
#


def is_int(val: object):
    return _isinstance(val, int)


def assert_int(val: object, name: str, message: Optional[str] = None):
    if not is_int(val):
        raise _assert_simple_type_fail_message("int", val, name, message)


def is_opt_int(val: object):
    return val is None or is_int(val)


def assert_opt_int(val: object, name: str, message: Optional[str] = None):
    if not is_opt_int(val):
        raise _assert_simple_type_fail_message("Optional[int]", val, name, message)


#
# float
#


def is_float(val: object):
    return _isinstance(val, float)


def assert_float(val: object, name: float, message: Optional[str] = None):
    if not is_float(val):
        raise _assert_simple_type_fail_message("float", val, name, message)


def is_opt_float(val: object):
    return val is None or is_float(val)


def assert_opt_float(val: object, name: float, message: Optional[str] = None):
    if not is_opt_float(val):
        raise _assert_simple_type_fail_message("Optional[float]", val, name, message)


#
# Numbers
#

Number = Union[float, int]


def is_number(val: object):
    return _isinstance(val, (float, int))


def assert_number(val: object, name: str, message: Optional[str] = None):
    if not is_number(val):
        raise _assert_simple_type_fail_message("Number", val, name, message)


def is_opt_number(val: object):
    return val is None or is_number(val)


def assert_opt_number(val: object, name: str, message: Optional[str] = None):
    if not is_opt_number(val):
        raise _assert_simple_type_fail_message("Optional[Number]", val, name, message)


def assert_positive(number: Number, name: str, message: Optional[str] = None):
    if number <= 0:
        _message = message

        if message is not None:
            _message = f"Expected variable {name} to be positive. Got {number}."

        raise _assert_true_fail_message(name, _message)


def assert_negative(number: Number, name: str, message: Optional[str] = None):
    if number >= 0:
        _message = message

        if message is not None:
            _message = f"Expected variable {name} to be negative. Got {number}."

        raise _assert_true_fail_message(name, _message)


def assert_non_positive(number: Number, name: str, message: Optional[str] = None):
    if number > 0:
        _message = message

        if message is not None:
            _message = f"Expected variable {name} to be non-positive. Got {number}."

        raise _assert_true_fail_message(name, _message)


def assert_non_negative(number: Number, name: str, message: Optional[str] = None):
    if number < 0:
        _message = message

        if message is not None:
            _message = f"Expected variable {name} to be non-negative. Got {number}."

        raise _assert_true_fail_message(name, _message)


#
# class
#


def is_instance(val: object, of: object):
    return _isinstance(val, of)


def assert_instance(val: object, of: object, name: str, message: Optional[str] = None):
    if not is_instance(val, of):
        raise _assert_simple_type_fail_message(type_name(of), val, name, message)


def is_opt_instance(val: object, of: object):
    return val is None or is_instance(val, of)


def assert_opt_instance(
    val: object, of: object, name: str, message: Optional[str] = None
):
    if not is_opt_instance(val, of):
        raise _assert_simple_type_fail_message(type_name(of), val, name, message)


#
# callable
#


def is_callable(val: object):
    return callable(val)


def assert_callable(val: object, name: str, message: Optional[str] = None):
    if not is_callable(val):
        raise _assert_simple_type_fail_message("Callable", val, name, message)


def is_opt_callable(val: object):
    return val is None or is_callable(val)


def assert_opt_callable(val: object, name: str, message: Optional[str] = None):
    if not is_opt_callable(val):
        raise _assert_simple_type_fail_message("Optional[Callable]", val, name, message)


#
# date
#


def is_date(val: object):
    return _isinstance(val, date)


def assert_date(val: object, name: str, message: Optional[str] = None):
    if not is_date(val):
        raise _assert_simple_type_fail_message("date", val, name, message)


def is_opt_date(val: object):
    return val is None or is_date(val)


def assert_opt_date(val: object, name: str, message: Optional[str] = None):
    if not is_opt_date(val):
        raise _assert_simple_type_fail_message("Optional[date]", val, name, message)


#
# datetime
#


def is_datetime(val: object):
    return _isinstance(val, datetime)


def assert_datetime(val: object, name: str, message: Optional[str] = None):
    if not is_datetime(val):
        raise _assert_simple_type_fail_message("datetime", val, name, message)


def is_opt_datetime(val: object):
    return val is None or is_datetime(val)


def assert_opt_datetime(val: object, name: str, message: Optional[str] = None):
    if not is_opt_datetime(val):
        raise _assert_simple_type_fail_message("Optional[datetime]", val, name, message)



from typing import Optional, Tuple, Union


class InvalidRuntype(Exception):
    ...


def _assert_simple_type_fail_message(
    expected_type_str: str, val: object, name: str, message: Optional[str] = None
):
    return InvalidRuntype(
        f"Expected variable {name} to be type {expected_type_str}. "
        + f"Got value {val!r} of type {type(val).__name__}. "
        + (message or "")
    )


TypeOrTuple = Union[type, Tuple[type, ...]]


def _isinstance(val: object, _type: TypeOrTuple):
    # Boolean values are instance of ints for python backwards compatability reasons
    # This behavior is not supported in this library
    if isinstance(val, bool) and (
        _type == int
        or (isinstance(_type, tuple) and int in _type and bool not in _type)
    ):
        return False

    return isinstance(val, _type)

def type_name(of: object):
    # Ehhhh
    try:
        return of.__name__.replace("typing.", "")
    except:
        return str(of).replace("typing.", "")



from typing import Optional

from rcheck.checks.basic_types import _assert_simple_type_fail_message
from rcheck.checks.shared import InvalidRuntype, _isinstance, type_name
from collections.abc import Sequence

#
# list
#
def _print_list_index(val, index):
    val_repr = repr(val)

    if len(val_repr) > 100:
        start_idx = index - 1 if index > 0 else 0
        end = index + 2
        
        print_repr = repr(val[start_idx:end])[1:-1]
        print_repr = f"[ ..., {print_repr}, ... ]"

        pre_spaces = " " * 6

    else:
        start_idx = 0
        print_repr = val_repr
        pre_spaces = ""

    spaces = pre_spaces + " " * (len(repr(val[start_idx:(index + 1)])) - 2)    
    cursor = spaces + "↑"   
    cursor += "\n" + spaces + f"| got type {type_name(type(val[index]))} at index {index}"

    return print_repr + "\n" + cursor

def _assert_list_type_fail_message(
    expected_type_str: str,
    val: object,
    name: str,
    failure_index: int,
    message: Optional[str] = None,
    list_name = "list"
):
    return InvalidRuntype(
        f"Expected items of the {list_name} {name} to be of type {expected_type_str}. \n\n"
        + _print_list_index(val, failure_index)
        + (f"\n\n{message}" if message is not None else "")
    )


def _is_iterable_of(iterable, sub_type):
    for idx, item in enumerate(iterable):
        if not _isinstance(item, sub_type):
            return False, idx

    return True, -1


def is_list(val: object, of: Optional[type] = None):
    if not _isinstance(val, list):
        return False

    return True if of is None else _is_iterable_of(val, of)[0]


def assert_list(
    val: object,
    name: str,
    message: Optional[str] = None,
    *,
    of: Optional[type] = None,
):
    if not _isinstance(val, list):
        raise _assert_simple_type_fail_message("list", val, name, message)

    if of is None:
        return

    items_pass_check, failed_idx = _is_iterable_of(val, of)
    if not items_pass_check:
        raise _assert_list_type_fail_message(
            type_name(of), val, name, failed_idx, message
        )


def _is_iterable_of_opt(iterable, sub_type):
    for idx, item in enumerate(iterable):
        if item is not None and not _isinstance(item, sub_type):
            return False, idx

    return True, -1


def is_list_of_opt(val: object, of: type):
    if not _isinstance(val, list):
        return False

    return _is_iterable_of_opt(val, of)[0]


def assert_list_of_opt(
    val: object,
    of: type,
    name: str,
    message: Optional[str] = None,
):
    if not _isinstance(val, list):
        raise _assert_simple_type_fail_message("list", val, name, message)

    items_pass_check, failed_idx = _is_iterable_of_opt(val, of)
    if not items_pass_check:
        raise _assert_list_type_fail_message(
            type_name(of), val, name, failed_idx, message
        )


def is_opt_list(val: object, of: Optional[type] = None):
    return val is None or is_list(val, of)


def assert_opt_list(
    val: object,
    name: str,
    message: Optional[str] = None,
    *,
    of: Optional[type] = None,
):
    if val is None:
        return

    if not _isinstance(val, list):
        raise _assert_simple_type_fail_message("list", val, name, message)

    if of is None:
        return

    items_pass_check, failed_idx = _is_iterable_of(val, of)
    if not items_pass_check:
        raise _assert_list_type_fail_message(
            type_name(of), val, name, failed_idx, message, list_name="optional list"
        )


#
# dict
#

def _assert_dict_type_fail_message(
    expected_type_str: str,
    val: object,
    name: str,
    failure_index: int,
    message: Optional[str] = None,
    dict_name = "dict"
):
    return InvalidRuntype(
        f"Expected items of the {list_name} {name} to be of type {expected_type_str}. \n\n"
        # + _print_list_index(val, failure_index)
        + (f"\n\n{message}" if message is not None else "")
    )


def assert_dict(
    val: object,
    name: str,
    message: Optional[str] = None,
    *,
    key_of: Optional[type] = None,
    val_of: Optional[type] = None,
):
    if not _isinstance(val, dict):
        raise _assert_simple_type_fail_message("dict", val, name, message)

    if key_of is not None:
        items_pass_check, failed_idx = _is_iterable_of(val.keys(), key_of)
        if not items_pass_check:
            raise _assert_list_type_fail_message(
                type_name(key_of), val, name, failed_idx, message
            )

    if val_of is not None:
        items_pass_check, failed_idx = _is_iterable_of(val.values(), val_of)
        if not items_pass_check:
            raise _assert_list_type_fail_message(
                type_name(val_of), val, name, failed_idx, message
            )

#
# sequences
#
def assert_sequence(
    val: object,
    name: str,
    message: Optional[str] = None,
    *,
    of: Optional[type] = None,
):
    if not _isinstance(val, Sequence):
        raise _assert_simple_type_fail_message("Sequence", val, name, message)

    if of is None:
        return

    items_pass_check, failed_idx = _is_iterable_of(val, of)
    if not items_pass_check:
        raise _assert_list_type_fail_message(
            type_name(of), val, name, failed_idx, message
        )
        


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rcheck', 'rcheck.checks']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'rcheck',
    'version': '0.0.7',
    'description': 'Runtime type checking',
    'long_description': '# `rcheck`\n\nRuntime type checking.\n\n---\n\nFunctions that return a boolean:\n* `is_<type>` such as `is_int`, `is_bool`, etc...\n* `is_opt_<type>` such as `is_opt_int`, `is_opt_bool`, etc.. for optional types\n\nFunctions that raise exceptions:\n* `assert_<type>` such as `assert_int`, `assert_bool`, etc...\n* `assert_opt_<type>` such as `assert_opt_int`, `assert_opt_bool`, etc... for optional types\n',
    'author': 'Alex Rudolph',
    'author_email': 'alex3rudolph@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.github.com/alrudolph/rcheck',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
